import localsolver.*;
import java.io.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Scanner;

public class LocationRoutingProblem {
    // LocalSolver
    private final LocalSolver localsolver;

    // Number of customers
    private int nbCustomers;

    // Customers coordinates
    private int[] xCustomers;
    private int[] yCustomers;

    // Customers demands
    private int[] demandsData;

    // Number of depots
    private int nbDepots;

    // Depots coordinates
    private int[] xDepots;
    private int[] yDepots;

    // Capacity of depots
    private int[] depotsCapacity;

    // Cost of opening a depot
    private double[] openingDepotsCost;

    // Number of trucks
    private int nbTrucks;

    // Capacity of trucks
    private int truckCapacity;

    // Cost of opening a route
    private int openingRouteCost;

    // Is the sequence used ?
    private LSExpression[] sequenceUsed;

    // What is the depot of the sequence ?
    private LSExpression[] associatedDepot;

    // Distance matrixes
    private double[][] distMatrixData;
    private double[][] distDepotsData;

    private int areCostDouble;

    // Decision variables
    private LSExpression[] customersSequences;
    private LSExpression[] depots;

    // Objective value
    private LSExpression totalCost;

    private LocationRoutingProblem(LocalSolver localsolver) {
        this.localsolver = localsolver;
    }

    private void solve(int limit) {
        // Declare the optimization model
        LSModel m = localsolver.getModel();

        int totalDemand = 0;
        for (int d : demandsData) {
            totalDemand += d;
        }
        int minNbTrucks = (int) Math.ceil(totalDemand / truckCapacity);
        int nbTrucks = (int) Math.ceil(1.5 * minNbTrucks);

        customersSequences = new LSExpression[nbTrucks];
        depots = new LSExpression[nbDepots];

        // A route is represented as a list containing the customers in the order they are visited
        for (int i = 0; i < nbTrucks; ++i) {
            customersSequences[i] = m.listVar(nbCustomers);
        }
        // All customers should be assigned to a route
        m.constraint(m.partition(customersSequences));
        // A depot is represented as a set containing the associated customersSequences
        for (int d = 0; d < nbDepots; ++d) {
            depots[d] = m.setVar(nbTrucks);
        }
        // All the customersSequences should be assigned to a depot
        m.constraint(m.partition(depots));

        LSExpression demands = m.array(demandsData);
        LSExpression distMatrix = m.array(distMatrixData);
        LSExpression distDepots = m.array(distDepotsData);
        sequenceUsed = new LSExpression[nbTrucks];
        LSExpression[] routeCosts = new LSExpression[nbTrucks];
        LSExpression[] distRoutes = new LSExpression[nbTrucks];
        associatedDepot = new LSExpression[nbTrucks];
        LSExpression quantityServed = m.array();

        for (int r = 0; r < nbTrucks; ++r) {
            LSExpression sequence = customersSequences[r];
            LSExpression c = m.count(sequence);

            // A sequence is used if it serves at least one customer
            sequenceUsed[r] = m.gt(c, 0);
            // The "find" function gets the depot that is assigned to the sequence
            associatedDepot[r] = m.find(m.array(depots), r);

            LSExpression demandLambda = m.lambdaFunction(j -> m.at(demands, j));
            quantityServed.addOperand(m.sum(sequence, demandLambda));
            // The quantity needed in each sequence must not exceed the vehicle capacity
            m.constraint(m.leq(m.at(quantityServed, r), truckCapacity));

            LSExpression distLambda = m
                .lambdaFunction(i -> m.at(distMatrix, m.at(sequence, i), m.at(sequence, m.sum(i, 1))));

            distRoutes[r] = m.sum(m.sum(m.range(0, m.sub(c, 1)), distLambda),
                m.iif(sequenceUsed[r], m.sum(m.at(distDepots, m.at(sequence, 0), associatedDepot[r]),
                    m.at(distDepots, m.at(sequence, m.sub(c, 1)), associatedDepot[r])), 0));
            // The sequence cost is the sum of the opening cost and the sequence length
            routeCosts[r] = m.sum(distRoutes[r], m.prod(openingRouteCost, sequenceUsed[r]));
        }

        LSExpression[] depotCost = new LSExpression[nbDepots];
        for (int d = 0; d < nbDepots; ++d) {
            // A depot is open if at least a sequence starts from there
            depotCost[d] = m.prod(openingDepotsCost[d], m.gt(m.count(depots[d]), 0));

            LSExpression depotLambda = m.lambdaFunction(r -> m.at(quantityServed, r));
            LSExpression depotQuantity = m.sum(depots[d], depotLambda);
            // The total demand served by a depot must not exceed its capacity
            m.constraint(m.leq(depotQuantity, depotsCapacity[d]));
        }
        LSExpression depotsCost = m.sum(depotCost);
        LSExpression routingCost = m.sum(routeCosts);
        totalCost = m.sum(routingCost, depotsCost);

        m.minimize(totalCost);
        m.close();

        localsolver.getParam().setTimeLimit(limit);
        localsolver.solve();
    }

    /* Write the solution in a file */
    private void writeSolution(String infile, String outfile) throws IOException {
        try (PrintWriter output = new PrintWriter(outfile)) {
            output.println("File name: " + infile + "; total cost = " + totalCost.getDoubleValue());
            for (int r = 0; r < nbTrucks; ++r) {
                if (sequenceUsed[r].getIntValue() != 0) {
                    output.print("Route " + r + ", assigned to depot " + associatedDepot[r].getIntValue() + " : ");
                    LSCollection customersCollection = customersSequences[r].getCollectionValue();
                    for (int i = 0; i < customersCollection.count(); ++i) {
                        output.print(customersCollection.get(i) + " ");
                    }
                    output.println();
                }
            }
        } catch (FileNotFoundException e) {
            System.out.println("An error occurred.");
            e.printStackTrace();
        }
    }

    private void readInstanceLrp(String fileName) throws IOException {
        try (Scanner infile = new Scanner(new File(fileName))) {
            nbCustomers = infile.nextInt();
            xCustomers = new int[nbCustomers];
            yCustomers = new int[nbCustomers];
            demandsData = new int[nbCustomers];
            distMatrixData = new double[nbCustomers][];
            distDepotsData = new double[nbCustomers][];
            nbDepots = infile.nextInt();
            xDepots = new int[nbDepots];
            yDepots = new int[nbDepots];
            depotsCapacity = new int[nbDepots];
            openingDepotsCost = new double[nbDepots];

            for (int i = 0; i < nbDepots; ++i) {
                xDepots[i] = infile.nextInt();
                yDepots[i] = infile.nextInt();
            }
            for (int i = 0; i < nbCustomers; ++i) {
                xCustomers[i] = infile.nextInt();
                yCustomers[i] = infile.nextInt();
            }
            truckCapacity = infile.nextInt();
            for (int i = 0; i < nbDepots; ++i) {
                depotsCapacity[i] = infile.nextInt();
            }
            for (int i = 0; i < nbCustomers; ++i) {
                demandsData[i] = infile.nextInt();
            }
            double[] tempOpeningCostDepots = new double[nbDepots];
            for (int i = 0; i < nbDepots; ++i) {
                if (infile.hasNext()) {
                    tempOpeningCostDepots[i] = Double.parseDouble(infile.next());
                } else if (infile.hasNextInt()) {
                    tempOpeningCostDepots[i] = infile.nextInt();
                }
            }
            int tempOpeningCostRoute = infile.nextInt();
            areCostDouble = infile.nextInt();
            if (areCostDouble == 1) {
                openingRouteCost = tempOpeningCostRoute;
                for (int i = 0; i < tempOpeningCostDepots.length; ++i) {
                    openingDepotsCost[i] = tempOpeningCostDepots[i];
                }
            } else {
                openingRouteCost = Math.round(tempOpeningCostRoute);
                for (int i = 0; i < tempOpeningCostDepots.length; ++i) {
                    openingDepotsCost[i] = Math.round(tempOpeningCostDepots[i]);
                }
            }
            computeDistanceMatrix();
        } catch (FileNotFoundException e) {
            System.out.println("An error occurred.");
            e.printStackTrace();
        }
    }

    void computeDistanceMatrix() {
        for (int i = 0; i < nbCustomers; ++i) {
            distMatrixData[i] = new double[nbCustomers];
            for (int j = 0; j < nbCustomers; ++j) {
                distMatrixData[i][j] = computeDist(xCustomers[i], xCustomers[j], yCustomers[i], yCustomers[j],
                    areCostDouble);
            }
            distDepotsData[i] = new double[nbDepots];
            for (int d = 0; d < nbDepots; ++d) {
                distDepotsData[i][d] = computeDist(xCustomers[i], xDepots[d], yCustomers[i], yDepots[d], areCostDouble);
            }
        }
    }

    private double computeDist(int xi, int xj, int yi, int yj, int areCostDouble) {
        double dist = Math.sqrt(Math.pow(xi - xj, 2) + Math.pow(yi - yj, 2));
        if (areCostDouble == 0) {
            dist = Math.ceil(100 * dist);
        }
        return dist;
    }

    public static void main(String[] args) {
        if (args.length < 1) {
            System.err.println("Usage: java LocationRoutingProblem inputFile [outputFile] [timeLimit]");
            System.exit(1);
        }

        try (LocalSolver localsolver = new LocalSolver()) {
            String instanceFile = args[0];
            String strOutfile = args.length > 1 ? args[1] : null;
            String strTimeLimit = args.length > 2 ? args[2] : "20";

            LocationRoutingProblem model = new LocationRoutingProblem(localsolver);
            model.readInstanceLrp(instanceFile);
            model.solve(Integer.parseInt(strTimeLimit));
            if (strOutfile != null)
                model.writeSolution(instanceFile, strOutfile);
        } catch (Exception ex) {
            System.err.println(ex);
            ex.printStackTrace();
            System.exit(1);
        }
    }
}
