use io;

/* Read instance data */
function input() {

    local usage = "Usage : localsolver location_routing_problem.lsp "
            + "inFileName=inputFile [solFileName=outputFile] [lsTimeLimit=timeLimit]";
    if (inFileName == nil) throw usage;

    readInputLrp();
    computeDistances();

    minNbTrucks = ceil(sum[c in 0...nbCustomers](demands[c]) / truckCapacity);
    nbTrucks = ceil(1.5 * minNbTrucks);
}

/* Declare the optimization model */
function model() {
    // A route is represented as a list containing the customers in the order they are
    // visited
    customersSequences[r in 0...nbTrucks] <- list(nbCustomers);
    // A depot is represented as a set containing the associated customersSequences
    depots[d in 0...nbDepots] <- set(nbTrucks);
    
    // All customers should be assigned to a route
    constraint partition(customersSequences);
    // All the customersSequences should be assigned to a depot
    constraint partition(depots);
    
    for [r in 0...nbTrucks] {
        local sequence <- customersSequences[r];
        local c <- count(sequence);

        quantityServed[r] <- sum(sequence, j => demands[j]);
        // The quantity needed in each route must not exceed the vehicle capacity
        constraint quantityServed[r] <= truckCapacity;
        // A route is used if it serves at least one customer
        sequenceUsed[r] <- c > 0;

        // The "find" function gets the depot that is assigned to the route
        associatedDepot[r] <- find(depots, r);
        
        sequenceLength[r] <- sum(0...c-1, i => distCustomers[sequence[i]][sequence[i + 1]])
                + (sequenceUsed[r] ? distCustomersDepots[sequence[0]][associatedDepot[r]]
                + distCustomersDepots[sequence[c - 1]][associatedDepot[r]] : 0);
        // The route cost is the sum of the opening cost and the route length
        routeCost[r] <- sequenceLength[r] + sequenceUsed[r] * openingRouteCost;
    }

    for [d in 0...nbDepots] {
        // The total demand served by a depot must not exceed its capacity
        constraint sum(depots[d], i => quantityServed[i]) <= depotsCapacity[d];
        // A depot is open if at least a route starts from there
        isDepotOpen[d] <- count(depots[d]) > 0;
    }

    depotsCost <- sum[d in 0...nbDepots](isDepotOpen[d] * openingCostDepots[d]);
    routingCost <- sum[r in 0...nbTrucks](routeCost[r]);

    totalCost <- routingCost + depotsCost;
    minimize totalCost;
}

function param() {
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}

/* Write the solution in a file */
function output() {
    if (solFileName == nil) return;
    local resFile = io.openWrite(solFileName);
    resFile.println("File name: ", inFileName + "; totalCost = " + totalCost.value);
    for [r in 0...nbTrucks] {
        if (sequenceUsed[r].value) {
            resFile.print("Route ", r, ", assigned to depot ", associatedDepot[r].value, ": ");
            for [customer in customersSequences[r].value] {
                resFile.print(customer, " ");
            }
            resFile.println();
        }
    }
    resFile.close();
}

function readInputLrp() {
    if (inFileName.endsWith(".dat")) readInputDat();
    else throw "Unknown file format";
}

function readInputDat() {
    local inFile = io.openRead(inFileName);

    nbCustomers = inFile.readInt();
    nbDepots = inFile.readInt();

    coordinatesDepots[d in 0...nbDepots][l in 0...2] = inFile.readDouble();
    coordinatesCustomers[c in 0...nbCustomers][l in 0...2] = inFile.readDouble();

    truckCapacity = inFile.readInt();
    depotsCapacity[d in 0...nbDepots] = inFile.readInt();

    demands[c in 0...nbCustomers] = inFile.readInt();

    local tempOpeningCostDepots[d in 0...nbDepots] = inFile.readDouble();
    local tempOpeningCostRoute = inFile.readDouble();

    areCostsDouble = inFile.readInt();

    if (areCostsDouble == 1) {
        openingCostDepots[d in 0...nbDepots] = tempOpeningCostDepots[d];
        openingRouteCost = tempOpeningCostRoute;
    } else {
        openingCostDepots[d in 0...nbDepots] = round(tempOpeningCostDepots[d]);
        openingRouteCost = round(tempOpeningCostRoute);
    }
}

function computeDistances() {
    local tempDistanceCustomers[c0 in 0...nbCustomers][c1 in 0...nbCustomers] =
            sqrt(pow((coordinatesCustomers[c0][0] - coordinatesCustomers[c1][0]), 2)
            + pow((coordinatesCustomers[c0][1] - coordinatesCustomers[c1][1]), 2));
    local tempDistanceCustomersDepots[c in 0...nbCustomers][d in 0...nbDepots] =
            sqrt(pow((coordinatesCustomers[c][0] - coordinatesDepots[d][0]), 2)
            + pow((coordinatesCustomers[c][1] - coordinatesDepots[d][1]), 2));

    if (areCostsDouble == 1) {
        distCustomers[c0 in 0...nbCustomers][c1 in 0...nbCustomers] =
                tempDistanceCustomers[c0][c1];
        distCustomersDepots[c in 0...nbCustomers][d in 0...nbDepots] =
                tempDistanceCustomersDepots[c][d];
    } else {
        distCustomers[c0 in 0...nbCustomers][c1 in 0...nbCustomers] =
                ceil(100 * tempDistanceCustomers[c0][c1]);
        distCustomersDepots[c in 0...nbCustomers][d in 0...nbDepots] =
                ceil(100 * tempDistanceCustomersDepots[c][d]);
    }
}